from typing import Any, Dict, List, Sequence

from ava_llm.conversations.messages.ai import AIMessage
from ava_llm.conversations.messages.base import BaseMessage
from ava_llm.conversations.messages.human import HumanMessage


def get_buffer_string(
    messages: Sequence[BaseMessage], human_prefix: str = "Human", ai_prefix: str = "AI"
) -> str:
    """Convert a sequence of Messages to strings and concatenate them into one string.

    Args:
        messages: Messages to be converted to strings.
        human_prefix: The prefix to prepend to contents of HumanMessages.
        ai_prefix: THe prefix to prepend to contents of AIMessages.

    Returns:
        A single string concatenation of all input messages.

    """
    string_messages = []
    for m in messages:
        if isinstance(m, HumanMessage):
            role = human_prefix
        elif isinstance(m, AIMessage):
            role = ai_prefix
        else:
            raise ValueError(f"Got unsupported message type: {m}")
        message = f"{role}: {m.content}"
        string_messages.append(message)

    return "\n".join(string_messages)


def get_prompt_input_key(inputs: Dict[str, Any]) -> str:
    """
    Get the prompt input key.

    Args:
        inputs: Dict[str, Any]

    Returns:
        A prompt input key.
    """
    if len(inputs) != 1:
        raise ValueError(f"Exactly one input key expected, got {len(inputs)}")
    return next(iter(inputs))



from groq import Groq
from dotenv import load_dotenv
import os

load_dotenv()

meta_key = os.getenv("META_KEY")

client = Groq(api_key=meta_key)


class Meta:
    def __init__(self, model, temperature=None):
        self.model = model or "llama3-8b-8192"
        self.temperature = temperature

    def invoke(self, prompt, temperature=None, force_json=False, **kwargs):
        temperature = temperature if temperature is not None else self.temperature
        response = self._generate(prompt, temperature, force_json=force_json, **kwargs)
        response_dict = self._response_to_dict(response)
        return response_dict

    async def ainvoke(self, prompt, temperature=None, **kwargs):
        raise NotImplementedError(
            "Asynchronous invocation is not supported for the llama model"
        )

    def stream(self, prompt, temperature=None, **kwargs):
        temperature = temperature if temperature is not None else self.temperature
        response = self._generate(prompt, temperature, stream=True, **kwargs)
        for chunk in response:
            yield chunk.choices[0].delta.content

    def _response_to_dict(self, response):
        # metadata dict
        metadata_dict = {}
        metadata_dict["input_tokens"] = response.usage.prompt_tokens
        metadata_dict["output_tokens"] = response.usage.completion_tokens
        metadata_dict["total_tokens"] = response.usage.total_tokens

        # response dict
        response_dict = {}
        response_dict["content"] = response.choices[0].message.content
        response_dict["response_metadata"] = {
            "token_usage": metadata_dict,
            "model": self.model,
        }
        return response_dict

    def _generate(
        self, prompt, temperature=1.0, stream=False, force_json=False, **kwargs
    ):
        if force_json:
            kwargs["response_format"] = {"type": "json_object"}

        prompt = self._format_prompt(prompt)

        response = client.chat.completions.create(
            model=self.model,
            messages=prompt,
            temperature=temperature,
            stream=stream,
            **kwargs
        )
        return response

    async def _agenerate(self, prompt, temperature=1.0, stream=False, **kwargs):
        raise NotImplementedError(
            "Asynchronous generation is not supported for the llama model"
        )

    def _format_prompt(self, prompt):
        if isinstance(prompt, str):
            return [{"role": "user", "content": prompt}]
        elif isinstance(prompt, list):
            return prompt
        else:
            raise TypeError("Unexpected prompt type")



from typing import Any, List
from ava_llm.conversations.chat_memory import BaseChatMemory
from ava_llm.conversations.messages.base import BaseMessage
from ava_llm.utils.utils import get_buffer_string


class ConversationBufferMemory(BaseChatMemory):
    human_prefix: str = "Human"
    ai_prefix: str = "AI"

    @property
    def buffer(self) -> Any:
        """String buffer of memory."""
        return self.buffer_as_messages if self.return_messages else self.buffer_as_str

    def _buffer_as_str(self, messages: List[BaseMessage]) -> str:
        return get_buffer_string(
            messages,
            human_prefix=self.human_prefix,
            ai_prefix=self.ai_prefix,
        )

    @property
    def buffer_as_str(self) -> str:
        """Exposes the buffer as a string in case return_messages is True."""
        return self._buffer_as_str(self.chat_memory.messages)

    @property
    def buffer_as_messages(self) -> List[BaseMessage]:
        """Exposes the buffer as a list of messages in case return_messages is False."""
        return self.chat_memory.messages



from ava_llm.prompts.base import PromptTemplate


class CustomTemplate(PromptTemplate):
    template = """{text}"""


class ConversationTemplate(PromptTemplate):
    template = """The following is a friendly conversation between a human and an AI. The AI is talkative and provides lots of specific details from its context. If the AI does not know the answer to a question, it truthfully says it does not know.
Current conversation:
{history}
Human: {input}
AI:
"""



from typing import Any, Dict, List, Optional, Union

class BaseMessage:
    """Base abstract Message class."""

    content: Union[str, List[Union[str, Dict]]]
    type: str
    id: Optional[str] = None

    def __init__(
        self, content: Union[str, List[Union[str, Dict]]], **kwargs: Any
    ) -> None:
        """Pass in content as positional arg."""
        self.content = content



from abc import ABC, abstractmethod
from typing import List, Sequence

from ava_llm.conversations.messages.base import BaseMessage
from ava_llm.utils.utils import get_buffer_string


class BaseChatMessageHistory(ABC):
    messages: List[BaseMessage]

    def add_message(self, message: BaseMessage) -> None:
        if type(self).add_messages != BaseChatMessageHistory.add_messages:
            self.add_messages([message])
        else:
            raise NotImplementedError(
                "add_message is not implemented for this class. "
                "Please implement add_message or add_messages."
            )

    def add_messages(self, messages: Sequence[BaseMessage]) -> None:
        for message in messages:
            self.add_message(message)

    @abstractmethod
    def clear(self) -> None:
        """Remove all messages from the store"""

    def __str__(self) -> str:
        return get_buffer_string(self.messages)


class InMemoryChatMessageHistory(BaseChatMessageHistory):
    messages: List[BaseMessage]

    def __init__(self) -> None:
        self.messages = []

    def add_message(self, message: BaseMessage) -> None:
        self.messages.append(message)

    def clear(self) -> None:
        self.messages = []



from setuptools import setup, find_packages

VERSION = "0.0.23"
DESCRIPTION = "A simple llm package."

# Setting up
setup(
    name="ava_llm",
    version=VERSION,
    description=DESCRIPTION,
    packages=find_packages(),
    install_requires=["openai", "groq", "python-dotenv"],
    keywords=["python"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from typing import Any
from ava_llm.conversations.chat_history import (
    BaseChatMessageHistory,
    InMemoryChatMessageHistory,
)
from ava_llm.prompts.base import PromptTemplate
from ava_llm.prompts.template import ConversationTemplate
import uuid
from itertools import tee


class LLM:
    def __init__(self, model=None, handler=None, temperature=None) -> None:
        self.model = model or "gpt-3.5-turbo"
        self.handler = handler or self._get_handler()
        self.temperature = temperature if not None else self.handler.temperature

    def invoke(self, prompt, temperature=None, force_json=False, **kwargs):
        temperature = temperature if temperature is not None else self.temperature
        response = self.handler.invoke(
            prompt, temperature=temperature, force_json=force_json, **kwargs
        )
        return self._format_response(response)

    async def ainvoke(self, prompt, temperature=None, **kwargs):
        temperature = temperature if temperature is not None else self.temperature
        response = await self.handler.ainvoke(prompt, temperature=temperature, **kwargs)
        return self._format_response(response)

    def stream(self, prompt, **kwargs):
        for chunk in self.handler.stream(prompt, **kwargs):
            if chunk:
                if hasattr(chunk, "content"):
                    yield chunk
                elif isinstance(chunk, str):
                    yield chunk
                else:
                    raise TypeError("Unexpected response type")

    def _format_response(self, response):

        # langchain response
        if hasattr(response, "content") and hasattr(response, "response_metadata"):
            return LLMResponse(response.content, response.response_metadata)

        # this library response
        if "content" in response and "response_metadata" in response:
            return LLMResponse(response["content"], response["response_metadata"])

        raise ValueError("Unexpected response type")

    def _get_handler(self):
        if (
            self.model == "gpt-3.5-turbo"
            or self.model == "gpt-4-turbo"
            or self.model == "gpt-4"
        ):
            from ava_llm.handlers.openai import OpenAI

            return OpenAI(model=self.model)
        elif self.model == "llama3-70b-8192" or self.model == "llama3-8b-8192":
            from ava_llm.handlers.meta import Meta

            return Meta(model=self.model)
        else:
            raise ValueError("Unsupported model")


class LLMResponse:
    def __init__(self, content, metadata=None):
        self.content = content
        self.metadata = metadata

    def __repr__(self):
        return f"LLMResponse(content='{self.content}', metadata={self.metadata})"


class ConversationChain:
    llm: LLM = None
    memory: BaseChatMessageHistory = None
    prompt: PromptTemplate = None
    _guid = str(uuid.uuid4())

    def __init__(
        self,
        llm: LLM,
        memory: BaseChatMessageHistory = InMemoryChatMessageHistory(),
        prompt: PromptTemplate = ConversationTemplate(),
    ) -> None:
        self.llm = llm
        self.memory = memory
        self.prompt = prompt

    @property
    def guid(self) -> str:
        return self._guid

    def stream(self, message: str, verbose=True) -> Any:
        inputs = {"prompt": message}

        prompt = self.prompt.format(history=self.memory.buffer, input=message)

        chunks = []
        response, clean_response = tee(self.llm.stream(prompt))

        for chunk in response:
            chunks.append(chunk)
            if verbose:
                print(chunk, end="", flush=True)

        outputs = {"response": "".join(chunks)}
        self.memory.save_context(inputs, outputs)

        return clean_response



class PromptTemplate:
    def format(self, **kwargs):
        return self.template.format(**kwargs)


class PromptWithSystemInstructionTemplate:
    """Prompt template with system instruction.
    
    Args:
        template (str): The template string for the normal (user) prompt.
        system_template (str): The template string for the system instruction.
    """
    def format(self, **kwargs):
        system_prompt = self.system_template.format(**kwargs)
        user_prompt = self.template.format(**kwargs)
        return [{"role": "system", "content": system_prompt}, {"role": "user", "content": user_prompt}]



from abc import ABC, abstractmethod
from typing import Any, Dict


class BaseMemory(ABC):
    """Abstract base class for memory in Chains.

    Memory refers to state in Chains. Memory can be used to store information about
        past executions of a Chain and inject that information into the inputs of
        future executions of the Chain. For example, for conversational Chains Memory
        can be used to store conversations and automatically add them to future model
        prompts so that the model has the necessary context to respond coherently to
        the latest input.

    """

    @abstractmethod
    def save_context(self, inputs: Dict[str, Any], outputs: Dict[str, str]) -> None:
        """Save the context of this chain run to memory."""

    @abstractmethod
    def clear(self) -> None:
        """Clear memory contents."""
